# -*- coding:utf-8 -*-
# @Time    : 2020/04/12 上午12:00
# @Author  : ydf
"""
可自动实时调节线程数量的线程池。
不仅能扩大，还可缩小
"""

import atexit
import queue
import threading
import weakref

import time

from py_log import get_logger

_shutdown = False
_threads_queues = weakref.WeakKeyDictionary()
logger = get_logger(__name__)


def _python_exit():
    global _shutdown
    _shutdown = True
    items = list(_threads_queues.items())
    for t, q in items:
        q.put(None)
    for t, q in items:
        t.join()


atexit.register(_python_exit)


class _WorkItem(object):
    def __init__(self, fn, args, kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def run(self):
        try:
            self.fn(*self.args, **self.kwargs)
        except BaseException as exc:
            logger.error(f'函数 {self.fn.__name__} 中发生错误，错误原因是 {type(exc)} {exc} ')
            raise Exception(exc)

    def __str__(self):
        return f'{(self.fn.__name__, self.args, self.kwargs)}'


class CustomThreadPoolExecutor:
    def __init__(self, max_workers=None, thread_name_prefix=''):
        """
        最好需要兼容官方concurren.futures.ThreadPoolExecutor 和改版的BoundedThreadPoolExecutor，入参名字和个数保持了一致。
        :param max_workers:
        :param thread_name_prefix:
        """
        self._max_workers = max_workers or 4
        self._min_workers = 5
        self._thread_name_prefix = thread_name_prefix
        self.work_queue = queue.Queue(max_workers)
        # self._threads = set()
        self._threads = weakref.WeakSet()
        self._lock_compute_threads_free_count = threading.Lock()
        self.threads_free_count = 0
        self._shutdown = False
        self._shutdown_lock = threading.Lock()

    def set_min_workers(self, min_workers=10):
        self._min_workers = min_workers
        return self

    def change_threads_free_count(self, change_num):
        with self._lock_compute_threads_free_count:
            self.threads_free_count += change_num

    def submit(self, func, *args, **kwargs):
        with self._shutdown_lock:
            if self._shutdown:
                raise RuntimeError('不能添加新的任务到线程池')
        self.work_queue.put(_WorkItem(func, args, kwargs))
        self._adjust_thread_count()

    def _adjust_thread_count(self):
        # if len(self._threads) < self._threads_num:
        if self.threads_free_count < self._min_workers and len(self._threads) < self._max_workers:
            # t = threading.Thread(target=_work,
            #                      args=(self._work_queue,self))
            t = _CustomThread(self)
            t.setDaemon(True)
            t.start()
            self._threads.add(t)
            _threads_queues[t] = self.work_queue

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True
            self.work_queue.put(None)
        if wait:
            for t in self._threads:
                t.join()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


class _CustomThread(threading.Thread, ):
    def __init__(self, executorx: CustomThreadPoolExecutor):
        super().__init__()
        self._executorx = executorx
        self._run_times = 0

    # noinspection PyProtectedMember
    def _remove_thread(self, stop_resson=''):
        # noinspection PyUnresolvedReferences
        logger.info(stop_resson)
        self._executorx.change_threads_free_count(-1)
        self._executorx._threads.remove(self)
        _threads_queues.pop(self)

    # noinspection PyProtectedMember
    def run(self):
        # noinspection PyUnresolvedReferences

        self._executorx.change_threads_free_count(1)
        while True:
            try:
                work_item = self._executorx.work_queue.get(block=True, timeout=60)
            except queue.Empty:
                if self._executorx.threads_free_count > self._executorx._min_workers:
                    self._remove_thread(
                        f'当前线程超过60秒没有任务，线程池中不在工作状态中的线程数量是 {self._executorx.threads_free_count}，'
                        f'超过了指定的数量 {self._executorx._min_workers}')
                    break
                else:
                    continue

            # print(work_item)
            if work_item is not None:
                self._executorx.change_threads_free_count(-1)
                work_item.run()
                del work_item
                self._executorx.change_threads_free_count(1)
                continue
            if _shutdown or self._executorx._shutdown:
                self._executorx.work_queue.put(None)
                break


if __name__ == '__main__':
    pool = CustomThreadPoolExecutor(200)

    # pool = BoundedThreadPoolExecutor(200)


    def consume(x):
        time.sleep(2)
        print(x)


    def show_count():
        while 1:
            time.sleep(1)
            print(f'线程数量 {threading.active_count()}')


    threading.Thread(target=show_count).start()

    for i in range(1000):
        time.sleep(0.1)
        pool.submit(consume, i)



# -*- coding: utf-8 -*-
# @Time    : 2020/7/17 19:30
# @Author  : CC
# @Desc    : t_queue.py


def t_demo1():
    from leek import TaskPublisher, TaskConsumer

    # for zz in range(1, 11):
    #     TaskPublisher(queue_name='test1').pub({'a': zz, 'b': zz})
    #
    # def print_msg_dict2(a):
    #     print(f"t_demo1:{a}")

    # for zz in range(1, 11):
    #     TaskPublisher(queue_name='test1', fliter_rep=True).pub(a=zz, b=zz)
    #
    # def print_msg_dict1(a, b):
    #     print(f"t_demo1:{a},{b}")
    #     print(print_msg_dict1.meta)

    # for zz in range(1, 11):
    #     TaskPublisher(queue_name='test1', fliter_rep=True).pub(zz)
    #
    # def print_msg_dict3(a):
    #     print(f"t_demo1:{a}")
    #     print(print_msg_dict3.meta)

    def print_msg_dict4(a: dict):
        print(f"t_demo1:{a}")
        print(print_msg_dict4.meta)

    tasks = [dict(a=zz, b=zz) for zz in range(1, 11)]
    TaskPublisher(queue_name='test1', fliter_rep=True).pub_list(tasks, param_type='only')

    TaskConsumer(queue_name='test1', consuming_function=print_msg_dict4, fliter_rep=True).start()


if __name__ == '__main__':
    t_demo1()



# -*- coding:utf-8 -*-
# @Author cc
# @TIME 2020/5/10 00:42
import json
import os

import redis
from leek.utils import get_host_ip, get_day_str

from leek.base_queue import BaseQueue


class RedisQueue(BaseQueue):
    redis_conn_instance = {}

    def __init__(self, queue_name: str, fliter_rep: bool = False, priority: int = None, namespace='', **kwargs):
        """

        :param name: 队列名称
        :param priority: 优先级
        :param fliter_rep: 是否开始重复任务过滤
        :param namespace: 队列名命名空间
        :param kwargs: db连接动态参数
        """
        super().__init__(queue_name, fliter_rep, namespace=namespace, **kwargs)
        if namespace:
            self.queue_name = '%s:%s' % (namespace, queue_name)
        else:
            self.queue_name = queue_name
        self.priority = priority
        if self.priority is not None:
            self.queue_name = f"{self.priority}_{self.queue_name}"
        self.fliter_rep = fliter_rep
        self.key_sets = self.queue_name + ':filter'
        self.heartbeat_field = f"{self.queue_name}:heartbeat_{os.getpid()}_{get_host_ip()}_{get_day_str()}"
        self.un_ack_sets = f"unack_message:{self.queue_name}"
        self.heartbeat_key = f"heartbeat_check:{self.queue_name}"

        self._db = self._getconn(**kwargs)
        self._init_queues = [f"{4 - a}_{queue_name}" for a in range(0, 5)]

    def _getconn(self, **kwargs):
        cache_key = f"redis_conn_{kwargs.get('host')}_{kwargs.get('db')}"
        if cache_key in self.redis_conn_instance:
            self.__db = self.redis_conn_instance.get(cache_key)
        else:
            self.__db = redis.Redis(**kwargs)
            self.redis_conn_instance[cache_key] = self.__db
        return self.__db

    def getdb(self):
        return self.__db

    def qsize(self):
        return self.__db.llen(self.queue_name)

    def isempty(self):
        return self.qsize() == 0

    def put(self, item):
        if self.fliter_rep:
            if self.__db.sismember(self.key_sets, item) is False:
                self.__db.lpush(self.queue_name, item)
                self.__db.sadd(self.key_sets, item)
        else:
            self.__db.lpush(self.queue_name, item)

    def clear(self):
        self.__db.delete(self.queue_name)
        if self.key_sets:
            self.__db.delete(self.key_sets)
        self.__db.delete(self.un_ack_sets)
        self.__db.delete('dlq:'+self.queue_name)

    def get(self, block=False, timeout=None):
        if block:
            if self.priority is None:
                item = self.__db.brpop(self.queue_name, timeout=timeout)
            else:
                item = self.__db.brpop(self._init_queues, timeout=timeout)
        else:
            item = self.__db.rpop(self.queue_name)
        if item:
            item = item[1].decode('utf-8') if isinstance(item, tuple) else item.decode('utf-8')
        return item

    def check_has_customer(self, hash_value):
        return self.getdb().sismember(self.key_sets, hash_value)

    def add_customer_task(self, hash_value):
        self.getdb().sadd(self.key_sets, hash_value)

    def ack(self, message):
        self.getdb().srem(self.un_ack_sets, json.dumps(message) if isinstance(message, dict) else message)

    def un_ack(self, message):
        self.getdb().sadd(self.un_ack_sets, json.dumps(message) if isinstance(message, dict) else message)

    def dlq_re_consume(self):
        all_queues = [self.queue_name]
        if self.priority is not None:
            all_queues = self._init_queues
        for queue_name in all_queues:
            dlq_queue_name = f'dlq:{queue_name}'
            dlq_queue_length = self.getdb().llen(dlq_queue_name)
            for aa in range(0, dlq_queue_length):
                dql_message = self.getdb().rpop(dlq_queue_name)
                print(aa)
                self.getdb().lpush(queue_name, dql_message)


if __name__ == '__main__':
    redis_host = '127.0.0.1'
    redis_password = ''
    redis_port = 6379
    redis_db = 0
    r_queue = RedisQueue('test12', host=redis_host, port=redis_port, db=redis_db,
                         password=redis_password)
    for i in range(1, 10):
        r_queue.put('123456')
    print(r_queue.qsize())
    r_queue.put('456')
    print(r_queue.get())
    print(r_queue.qsize())
    print(r_queue.check_has_customer('123'))
    r_queue.un_ack('66666')
    r_queue.ack('123456')
    r_queue.dlq_re_consume()
    r_queue.clear()



# -*- coding:utf-8 -*-
# @Author cc
# @TIME 2020/5/13 11:30
import json

import time
import traceback

from kafka import KafkaConsumer, KafkaProducer
from leek import default_config

from leek.base_queue import BaseQueue

kafka_conn_instance = {}


class KafkaConsumerOwner(object):
    """
    使用Kafka—python的消费模块
    """

    def __init__(self, host, port, topic, groupid='default'):
        self.kafkatopic = topic
        self.groupid = groupid
        if ',' in host:
            self.bootstrap_servers = ','.join(
                [f'{host_pre}:{port}' if port and str(port) not in host_pre else host_pre for host_pre in
                 host.split(',')])
        else:
            self.bootstrap_servers = f'{host}:{port}' if port and str(port) not in host else host

        key = f"consumer_{self.kafkatopic}_{self.bootstrap_servers}"
        if kafka_conn_instance.get(key):
            self.consumer = kafka_conn_instance.get(key)
        else:
            if default_config.kafka_username and default_config.kafka_password:
                self.consumer = KafkaConsumer(self.kafkatopic,
                                              group_id=groupid,
                                              sasl_plain_username=default_config.kafka_username,
                                              sasl_plain_password=default_config.kafka_password,
                                              bootstrap_servers=self.bootstrap_servers, enable_auto_commit=True)
            else:
                self.consumer = KafkaConsumer(self.kafkatopic,
                                              group_id=groupid,
                                              bootstrap_servers=self.bootstrap_servers, enable_auto_commit=True)
            kafka_conn_instance[key] = self.consumer

    def get_msg(self):
        return self.consumer.poll(1000)


class KafkaProducerOwner(object):
    """
    使用kafka的生产模块
    """

    def __init__(self, host, port, topic):
        self.kafkatopic = topic
        if ',' in host:
            self.bootstrap_servers = ','.join(
                [f'{host_pre}:{port}' if port and str(port) not in host_pre else host_pre for host_pre in host.split(',')])
        else:
            self.bootstrap_servers = f'{host}:{port}' if port and str(port) not in host else host
        key = f"producer_{self.kafkatopic}_{self.bootstrap_servers}"
        if kafka_conn_instance.get(key):
            self.producer = kafka_conn_instance.get(key)
        else:
            if default_config.kafka_username and default_config.kafka_password:
                self.producer = KafkaProducer(sasl_plain_username=default_config.kafka_username,
                                              sasl_plain_password=default_config.kafka_password,
                                              bootstrap_servers=self.bootstrap_servers)
            else:
                self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)
            kafka_conn_instance[key] = self.producer

    def send_msg(self, params):
        parmas_message = json.dumps(params) if isinstance(params, dict) else params
        self.producer.send(self.kafkatopic, parmas_message.encode())


class KafkaQueue(BaseQueue):

    # noinspection PyBroadException
    def put(self, item):
        try:
            self.producer_k.send_msg(item)
        except Exception:
            traceback.print_exc()

    def getdb(self):
        super().getdb()

    def qsize(self):
        super().qsize()

    # noinspection PyBroadException
    def get(self, block=False, timeout=None):
        try:
            rs = self.consumer_k.consumer.poll(timeout_ms=1000)
            result = []
            for key in rs:
                for record in rs[key]:
                    result.append(json.loads(record.value.decode()))
        except Exception:
            traceback.print_exc()
            return None
        return result

    def _getconn(self, **kwargs):
        self.producer_k = KafkaProducerOwner(topic=self.queue_name, **kwargs)
        self.consumer_k = KafkaConsumerOwner(topic=self.queue_name, **kwargs)

    def isempty(self):
        super().isempty()

    def clear(self):
        self.producer_k.producer.flush()


if __name__ == '__main__':
    kafka_queue = KafkaQueue(queue_name='test_leek', host=default_config.kafka_host, port=default_config.kafka_port)
    while True:
        kafka_queue.put('123456789')
        time.sleep(3)
